"""

SetupTools implementation for all of the mirantis Metta packages

does not include the individual suites, which each act as their own packages.

"""
from setuptools import setup

if __name__ == "__main__":
    setup()
